/*

	Rikaichan
	Copyright (C) 2005-2017 Jonathan Zarate
	http://www.polarcloud.com/

	---

	Originally based on RikaiXUL 0.4 by Todd Rudick
	http://www.rikai.com/
	http://rikaixul.mozdev.org/

	---

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	---

	Please do not change or remove any of the copyrights or links to web pages
	when modifying any of the files.

*/

Components.utils.import('resource://gre/modules/Console.jsm');
Components.utils.import('chrome://rikaichan/content/prefs.jsm');


var rcxMain = {
	init: function() {
		window.addEventListener('load', this._onLoad, false);
	},

	onLoad: function() {
		window.removeEventListener('load', this._onLoad, false);
		window.addEventListener('unload', this._onUnload, false);

		let ks = document.getElementById('mailKeys') || document.getElementById('editorKeys');
		if (ks) {
			this.isTB = true;
		}
		else {
			ks = document.getElementById('mainKeyset') || document.getElementById('navKeys');
			this.isTB = false;
		}
		if (ks) {
			let prefs = new rcxPrefs();
			['toggle', 'lbar'].forEach(function(name) {
				let s = prefs.getString(name + '.key');
				if ((s.length) && (s != '(disabled)')) {

					let key = document.createElement('key');
					if (s.length > 1) key.setAttribute('keycode', 'VK_' + s.replace(' ', '_').toUpperCase());	// "Page Up" -> "VK_PAGE_UP"
						else key.setAttribute('key', s);
					key.setAttribute('modifiers', prefs.getString(name + '.mod'));
					key.setAttribute('command', 'rikaichan-' + name + '-command');
					ks.appendChild(key);
				}
			});
		}


		rcxConfig.load();
		rcxConfig.addObserver();

		this.addObserver();

		let prefs = new rcxPrefs();
		if (prefs.getBool('firsticon')) {
			prefs.setBool('firsticon', false);
			let nb = document.getElementById('nav-bar') || document.getElementById('mail-bar3');
			if (nb) {
				nb.insertItem('rikaichan-toggle-button');
				nb.setAttribute('currentset', nb.currentSet);
				document.persist(nb.id, 'currentset');
			}
		}

		if (this.isTB) {
			rcxTB.init();

			let nb = document.getElementById('composeToolbar2');
			if (nb) {
				if (prefs.getBool('firsticoncompose')) {
					prefs.setBool('firsticoncompose', false);
					nb.insertItem('rikaichan-toggle-button');
					nb.setAttribute('currentset', nb.currentSet);
					document.persist(nb.id, 'currentset');
				}
			}
		}
		else {
			gBrowser.tabContainer.addEventListener('TabSelect', rcxMain._onTabSelected, false);

			// enmode: 0=tab, 1=browser, 2=all, 3=always
			if (rcxConfig.enmode == 3) {
				this.enabledOnAll = true;
				this.onTabSelected();
			}
		}


		this.checkVersion();
	},

	_onLoad: function() {
		rcxMain.onLoad();
	},


	onUnload: function() {
		this.removeObserver();
		rcxConfig.removeObserver();

		if (this.isTB) {
			rcxTB.done();
		}
		else {
			gBrowser.tabContainer.removeEventListener('TabSelect', rcxMain._onTabSelected, false);
		}

		window.removeEventListener('unload', this._onUnload, false);
	},

	_onUnload: function() {
		rcxMain.onUnload();
	},


	onContentMessage: function(msg) {
		// TODO: allow setting index, return index
		if (msg.data.action == 'word-search') {
			let e = rcxData.wordSearch(msg.data.text);
			if (e != null) {
				e.html = rcxData.makeHtml(e);
			}
			return e;
		}

		// console.log('\nonContentMessage');
		// console.log('name=' + msg.name);
		// console.log('sync=' + msg.sync);
		// console.log('data=', msg.data);
		// console.log('target=', msg.target);
		// console.log('objects=', msg.objects);

		if (msg.data.action == 'translate') {
			let e = rcxData.translate(msg.data.text);
			if (e != null) {
				e.title = msg.data.text.substr(0, e.textLen).replace(/[\x00-\xff]/g, function (c) { return '&#' + c.charCodeAt(0) + ';' } );
				if (msg.data.text.length > e.textLen) e.title += '...';
				e.html = rcxData.makeHtml(e);
			}
			return e;
		}

		// TODO: allow setting index, return index
		if (msg.data.action == 'lookup-search') {
			return rcxData.lookupSearch(msg.data.text);
		}

		// TODO: need to remove
		if (msg.data.action == 'data-next') {
			rcxData.selectNext();
			return {};
		}

		// TODO: need to remove
		if (msg.data.action == 'data-select') {
			rcxData.select(msg.data.index);
			return {};
		}

		if (msg.data.action == 'copy') {
			this._copyToClipboard(msg.data.entries);
			return;
		}

		if (msg.data.action == 'save') {
			this._saveToFile(msg.data.entries);
			return;
		}

		if (msg.data.action == 'lookupbar-set-text') {
			rcxLookupBar.setText(msg.data.text);
			return;
		}

		// TODO: remove, unused
		if (msg.data.action == 'get-selected-return') {
			rcxLookupBar.setText(msg.data.text);
			return;
		}

		if (msg.data.action == 'inserted') {
			let b = msg.target;
			if ((typeof(b._rikaichan_) != 'undefined') && (b._rikaichan_.enabled)) {
				// console.log('msg inserted: re-enabling')
				this.sendAsyncMessage(b, { action: 'enable' });
				this.sendConfig(b);
			}
			else {
				// console.log('msg inserted: not enabled')
			}
			return;
		}

		if (msg.data.action == 'disabled') {
			// for TB
			rcxTB.disabled();
			return;
		}

		if (msg.data.action == 'paused') {
			// for TB
			rcxTB.disabled();
			return;
		}
	},

	_onContentMessage: function(msg) {
		return rcxMain.onContentMessage(msg);
	},

	sendAsyncMessage: function(browser, data) {
		// console.log('sendAsyncMessage data=', data);

		if (browser.messageManager) {
			browser.messageManager.sendAsyncMessage('rcx@polarcloud.com:msg', data);
		}
		else if (window.messageManager) {
			// Thunderbird/Compose: browser is actually an editor (use window.messageManager); there is only 1 in this window
			window.messageManager.broadcastAsyncMessage('rcx@polarcloud.com:msg', data);
		}
		else {
			console.log('ERROR: no messageManager');
		}
	},

	_makeConfig: function() {
		return {
			action: 'config',
			config: {
				popdelay: rcxConfig.popdelay,
				highlight: rcxConfig.highlight,
				title: rcxConfig.title,
				resizedoc: rcxConfig.resizedoc,
				nopopkeys: rcxConfig.nopopkeys,
				hidedef: rcxConfig.hidedef,
				usedpr: rcxConfig.usedpr,
				popdy: rcxConfig.popdy
			}
		};
	},

	resendConfig: function() {
		if (this.frameInserted) {
			window.messageManager.broadcastAsyncMessage('rcx@polarcloud.com:msg', this._makeConfig());
		}
	},

	sendConfig: function(browser) {
		if (typeof(browser._rikaichan_) != 'undefined') {
			this.sendAsyncMessage(browser, this._makeConfig());
		}
	},

	insertFrame: function(browser) {
		// console.log('insertFrame() browser=', browser);

		// Thunderbird/Compose: browser = editor; no messageManager (use window.messageManager)
		// let mm = browser.messageManager || window.messageManager;

		if (typeof(browser._rikaichan_) == 'undefined') {
			browser._rikaichan_ = { listening: false, enabled: false };
			// console.log('new browser._rikaichan_');
		}


		if (!this.frameInserted) {
			this.frameInserted = true;
			let wmm = window.messageManager;
			wmm.loadFrameScript('chrome://rikaichan/content/frame.js', true, false);
			wmm.addMessageListener('rcx@polarcloud.com:msg', this._onContentMessage);
		}

/*
		// Does not work correctly: If inserted into a new/blank tab, the frame script is unloaded
		// when a new content is loaded, the message listener is removed (?).
		if (!browser._rikaichan_.loaded) {
			browser._rikaichan_.loaded = true;
			mm.loadFrameScript('chrome://rikaichan/content/frame.js', true);

			console.log('frame loaded');
		}
		else {
			console.log('frame already loaded');
		}

		if (!browser._rikaichan_.listening) {
			browser._rikaichan_.listening = true;
			mm.addMessageListener('rcx@polarcloud.com:msg', this._onContentMessage);
		}
*/
	},

	enableOnBrowser: function(browser, forLookupBar) {
		if (!this.initDictionary()) return;

		this.insertFrame(browser);

		if (forLookupBar) {
			browser._rikaichan_.lookupBar = true;
			return;
		}

		if (!browser._rikaichan_.enabled) {
			browser._rikaichan_.enabled = true;

			let h = null;
			if ((rcxConfig.minihelp) && (!this.helpShown)) {
				this.helpShown = true;
				h = rcxFile.read('chrome://rikaichan/locale/minihelp.htm');
			}
			this.sendAsyncMessage(browser, { action: 'enable', help: h });
			this.sendConfig(browser);
		}
	},

	disableOnBrowser: function(browser, forLookupBar) {
		if (typeof(browser._rikaichan_) != 'undefined') {
			// Thunderbird/Compose: browser = editor; no messageManager (use window.messageManager)
			//	let mm = browser.messageManager || window.messageManager;

			if (forLookupBar) {
				browser._rikaichan_.lookupBar = false;
				this.sendAsyncMessage(browser, { action: 'hide' });
			}
			else {
				this.sendAsyncMessage(browser, { action: 'disable' });
				browser._rikaichan_.enabled = false;
			}
/*
			if ((!browser._rikaichan_.enabled) && (!browser._rikaichan_.lookupBar)) {
				mm.removeMessageListener('rcx@polarcloud.com:msg', this._onContentMessage);
				browser._rikaichan_.listening = false;
			}
*/
			// note: frame scripts cannot be unloaded
		}
	},

	disableAll: function() {
		for (let i = gBrowser.browsers.length - 1; i >= 0; --i) {
			this.disableOnBrowser(gBrowser.browsers[i]);
		}
	},

	onTabSelected: function() {
		// console.log('rcxMain.onTabSelected');
		// console.log('enmode=' + rcxConfig.enmode);

		let browser = this.getBrowser();

		// enmode: 0=tab, 1=browser, 2=all, 3=always
		if (rcxConfig.enmode > 0) {
			if (this.enabledOnAll) this.enableOnBrowser(browser);
				else this.disableOnBrowser(browser);
		}

		let on = ((browser._rikaichan_ != null) && (browser._rikaichan_.enabled));

		let bt = document.getElementById('rikaichan-toggle-button');
		if (bt) {
			bt.setAttribute('rcx-enabled', on);
			// console.log('rcx-enabled=', bt.getAttribute('rcx-enabled'));
		}

		let cm = document.getElementById('rikaichan-toggle-cm');
		if (cm) cm.setAttribute('checked', on);
	},

	_onTabSelected: function() {
		// see rcxData.loadConfig
		if ((rcxData.dicPath) && (!rcxData.dicPath.ready)) {
			rcxMain.delayedSelect = true;
		}
		else {
			rcxMain.onTabSelected();
		}
	},


	onToggleButton: function(event) {
		// console.log('onToggleButton');

		let browser = this.getBrowser();
		let on = ((typeof(browser._rikaichan_) != 'undefined') && (browser._rikaichan_.enabled));

		// console.log('onToggleButton: on=' + on);

		this.missingAsked = false;

		// enmode: 0=tab, 1=browser, 2=all, 3=always
		if (rcxConfig.enmode > 0) {
			this.enabledOnAll = !on;
			// console.log('onToggleButton: enabledOnAll=' + this.enabledOnAll);
		}
		else {
			if (on) this.disableOnBrowser(browser);
				else this.enableOnBrowser(browser);
		}

		this.onTabSelected();
	},

	savePrep: function(entries, clip) {
		// console.log('savePrep: entries=', entries)

		if ((!entries) || (entries.length == 0)) return null;

		let me, mk;
		if (clip) {
			me = rcxConfig.smaxce;
			mk = rcxConfig.smaxck;
		}
		else {
			me = rcxConfig.smaxfe;
			mk = rcxConfig.smaxfk;
		}

		if (!entries.fromLB) mk = 1;

		let text = '';
		for (let i = 0; i < entries.length; ++i) {
			let e = entries[i];
			if (e.kanji) {
				if (mk-- <= 0) continue
				text += rcxData.makeText(e, 1);
			}
			else {
				if (me <= 0) continue;
				text += rcxData.makeText(e, me);
				me -= e.data.length;
			}
		}

		if (rcxConfig.snlf == 1) text = text.replace(/\n/g, '\r\n');
			else if (rcxConfig.snlf == 2) text = text.replace(/\n/g, '\r');

		let sep = rcxConfig.ssep;
		switch (sep) {
		case 'Tab':
			sep = '\t';
			break;
		case 'Comma':
			sep = ',';
			break;
		case 'Space':
			sep = ' ';
			break;
		}
		if (sep != '\t') return text.replace(/\t/g, sep);

		return text;
	},

	copyToClipboard: function() {
		this.sendAsyncMessage(this.getBrowser(), { action: 'copy' });
	},

	_copyToClipboard: function(entries) {
		let text;
		if ((text = this.savePrep(entries, true)) != null) {
			Components.classes['@mozilla.org/widget/clipboardhelper;1']
				.getService(Components.interfaces.nsIClipboardHelper)
				.copyString(text);
			this.showText('Copied to clipboard.');
		}
	},

	saveToFile: function() {
		this.sendAsyncMessage(this.getBrowser(), { action: 'save' });
	},

	_saveToFile: function(entries) {
		let fname = rcxConfig.sfile;

		try {
			let text;
			if ((text = this.savePrep(entries, false)) == null) return;

			let fu = Components.utils.import('resource://gre/modules/FileUtils.jsm').FileUtils;
			let f = null;

			if (fname != '') {
				f = new fu.File(fname);
			}

			if ((fname == '') || ((f != null) && (f.exists()) && (f.isDirectory()))) {
				const nsIFilePicker = Components.interfaces.nsIFilePicker;
				let fp = Components.classes['@mozilla.org/filepicker;1'].createInstance(nsIFilePicker);

				fp.init(window, 'Save As...', nsIFilePicker.modeSave);
				fp.appendFilters(nsIFilePicker.filterAll | nsIFilePicker.filterText);
				fp.defaultString = 'rikaichan.txt';
				fp.defaultExtension = 'txt';
				if (fname != '') fp.displayDirectory = new fu.File(fname);

				let r = fp.show();
				if ((r != nsIFilePicker.returnOK) && (r != nsIFilePicker.returnReplace)) return;

				fname = fp.file.path;
				f = new fu.File(fname);
			}


			let exists = f.exists();
			let fos = fu.openFileOutputStream(f, fu.MODE_CREATE | fu.MODE_APPEND | fu.MODE_WRONLY);
			if ((!exists) && (rcxConfig.ubom) && (rcxConfig.sfcs == 'utf-8')) {
				let bom = '\xEF\xBB\xBF';
				fos.write(bom, bom.length);
			}

			// note: nsIConverterOutputStream always adds BOM for UTF-16

			let os = Components.classes['@mozilla.org/intl/converter-output-stream;1']
					.createInstance(Components.interfaces.nsIConverterOutputStream);
			os.init(fos, rcxConfig.sfcs, 0, 0x3F);	// unknown -> '?'
			os.writeString(text);
			os.close();
			fos.close();

			this.showText('Saved to: ' + fname);
		}
		catch (ex) {
			this.showText('Error while saving ' + fname + ': ' + ex);
		}
	},

	showText: function(text) {
		let browser = this.getBrowser();
		if ((browser) && (browser._rikaichan_)) {
			this.sendAsyncMessage(browser, { action: 'show', text: text });
		}
		else {
			this.alertBox(text);
		}
	},


	getBrowser: function() {
		if (this.isTB) {
			if (this.tabMail) return this.tabMail.getBrowserForSelectedTab();
			return document.getElementById('messagepane') || document.getElementById('content-frame');
			// Compose: content-frame = editor
		}
		else {
			return gBrowser.selectedBrowser;
		}
	},

	initDictionary: function() {
		if (rcxData.missing) {
			if (this.missingAsked) return false;
			this.missingAsked = true;
			if (confirm('A dictionary file was not found. Show the download page?')) {
				this.showDownloadPage();
			}
			return false;
		}
		try {
			rcxData.init();
		}
		catch (ex) {
			this.alertBox('Error: ' + ex);
			return false;
		}
		return true;
	},

	configPage: function() {
		window.openDialog('chrome://rikaichan/content/options.xul', '', 'chrome,centerscreen');
	},

	showDownloadPage: function() {
		const url = 'http://www.polarcloud.com/getrcx/?version=' + (this.version || '');
		try {
			if (this.isTB) {
				Components.classes['@mozilla.org/messenger;1'].createInstance()
					.QueryInterface(Components.interfaces.nsIMessenger)
					.launchExternalURL(url);
			}
			else {
				gBrowser.selectedTab = gBrowser.addTab(url);
			}
		}
		catch (ex) {
			this.alertBox('There was an error opening ' + url);
		}
	},

	checkVersion: function() {
		Components.utils.import('resource://gre/modules/AddonManager.jsm');
		AddonManager.getAddonByID('{0AA9101C-D3C1-4129-A9B7-D778C6A17F82}', function(addon) {
			rcxMain.version = addon.version;
			let prefs = new rcxPrefs();
			if (prefs.getString('version') != addon.version) {
				prefs.setString('version', addon.version);
				setTimeout(function() {
					rcxMain.showDownloadPage();
				}, 2000);
			}
		});
	},

	alertBox: function(msg) {
		alert('[Rikaichan v' + (this.version || '') + '] ' + msg);
	},

	rcxObserver: {
		observe: function(subject, topic, data) {
			if (topic != 'rikaichan') return;

			if (data == 'data-ready') {
				if (rcxMain.delayedSelect) {
					rcxMain.delayedSelect = false;
					rcxMain.onTabSelected();
				}
				return;
			}

			if (data == 'show-download') {
				rcxMain.showDownloadPage();
				return;
			}
		}
	},

	addObserver: function() {
		Components.classes["@mozilla.org/observer-service;1"]
			.getService(Components.interfaces.nsIObserverService)
			.addObserver(this.rcxObserver, 'rikaichan', false);
	},

	removeObserver: function() {
		Components.classes['@mozilla.org/observer-service;1']
			.getService(Components.interfaces.nsIObserverService)
			.removeObserver(this.rcxObserver, 'rikaichan');
	}
};


var rcxTB = {
	init: function() {
		let tm = document.getElementById('tabmail');
		if (tm) tm.registerTabMonitor(this.tabMonitor);

		// In compose window, setup some hooks to remove Rikaichan before saving/sending. Not ideal, but...

		if (typeof(GenericSendMessage) == 'function') {
			//	console.log('GenericSendMessage:');
			//	console.log(GenericSendMessage);
			this.oldGenericSendMessage = GenericSendMessage;
			GenericSendMessage = this._GenericSendMessage;
		}

		// This will not work correctly now in an asynchronous environment:
		// Components.classes['@mozilla.org/observer-service;1']
		// 	.getService(Components.interfaces.nsIObserverService)
		// 	.addObserver(this.tbObserver, 'mail:composeOnSend', false);
	},

	done: function() {
		let tm = document.getElementById('tabmail');
		if (tm) tm.unregisterTabMonitor(this.tabMonitor);

		// Components.classes['@mozilla.org/observer-service;1']
		// 	.getService(Components.interfaces.nsIObserverService)
		// 	.removeObserver(this.tbObserver, 'mail:composeOnSend');
	},

	_GenericSendMessage: function(...args) {
		//	console.log('GenericSendMessage called args=', args);
		if (rcxMain.frameInserted) {
			rcxTB.oldGenericSendMessageArgs = args;
			rcxTB.execGenericSendMessageAfterDisable = true;
			window.messageManager.broadcastAsyncMessage('rcx@polarcloud.com:msg', { action: 'pause' });
		}
		else {
			// console.log('!frameInserted');
			rcxTB.oldGenericSendMessage.apply(null, args);
		}
	},

	disabled: function() {
		if (this.execGenericSendMessageAfterDisable) {
			delete this.execGenericSendMessageAfterDisable;

			//	console.log('execGenericSendMessageAfterDisable');
			rcxTB.oldGenericSendMessage.apply(null, this.oldGenericSendMessageArgs);
			delete this.oldGenericSendMessageArgs;

			window.messageManager.broadcastAsyncMessage('rcx@polarcloud.com:msg', { action: 'unpause' });
		}
	},

/*
	tbObserver: {
		observe: function(subject, topic, data) {
			if (topic == 'mail:composeOnSend') {
				let e = window.content.document.getElementById('rikaichan-window');
				if (e) e.parentNode.removeChild(e);
				console.log('mail:composeOnSend');
			}
		},
	},
*/

	tabMonitor: {
		monitorName: 'rikaichan',

		onTabSwitched: function(aTab, aOldTab) {
			rcxMain.onTabSelected();
		},

		onTabTitleChanged: function(aTab) { },
		onTabOpened: function(aTab, aIsFirstTab, aWasCurrentTab) { },
		onTabClosing: function(aTab) { },
		onTabPersist: function(aTab) { },
		onTabRestored: function(aTab, aState, aIsFirstTab) { }
	}
};


var rcxLookupBar = {
	setText: function(text) {
		if (!document.getElementById('rikaichan-lbar').hidden) {
			let e = document.getElementById('rikaichan-lbar-text');
			if (e) {
				e.value = text;
				e.focus();
			}
		}
	},

	close: function() {
		document.getElementById('rikaichan-lbar').hidden = true;
		document.getElementById('rikaichan-lbar-text').value = '';
		rcxData.done();

		rcxMain.disableOnBrowser(rcxMain.getBrowser(), true);
	},

	toggle: function() {
		this.lbText = document.getElementById('rikaichan-lbar-text');
		let e = document.getElementById('rikaichan-lbar');

		if (e.hidden) {
			e.hidden = false;
			this.lbText.value = '';
			this.lbText.focus();

			let browser = rcxMain.getBrowser();
			rcxMain.enableOnBrowser(browser, true);
			if (rcxConfig.selinlb) {
				rcxMain.sendAsyncMessage(browser, { action: 'lookup', text: null, checkSelected: true });
			}
		}
		else if (!this.lbText.getAttribute("focused")) {
			this.lbText.focus();
		}
		else {
			this.close();
		}
	},

	onKeyPress: function(ev) {
		switch (ev.keyCode) {
		case 13:
			let browser = rcxMain.getBrowser();
			rcxMain.enableOnBrowser(browser, true);
			rcxMain.sendAsyncMessage(browser, { action: 'lookup', text: this.lbText.value, checkSelected: false });
			ev.stopPropagation();
			break;
		case 27:
			this.close();
			ev.stopPropagation();
			break;
		}
	},

	onSearchButton: function(ev) {
		let browser = rcxMain.getBrowser();
		rcxMain.enableOnBrowser(browser, true);
		rcxMain.sendAsyncMessage(browser, { action: 'lookup', text: this.lbText.value, checkSelected: true });
	},

	search: function(text) {
		let s = text.replace(/^\s+|\s+$/g, '');
		if (s == '') return;

		if ((this.lbLast == s) && (this.isVisible())) {
			rcxData.selectNext();
		}
		else {
			this.lbLast = s;
			rcxData.select(0);
		}

		if (!this.initDictionary()) {
			this.hidePopup();
		}
		else {
			let result;
			let html;
			if ((s.search(/^:/) != -1) || (s.search(/^([^\u3000-\uFFFF]+)$/) != -1)) {
				// ":word"  = force a text search of "word"
				result = rcxData.textSearch(s.replace(/^:/, ''));
			}
			else {
				result = rcxData.wordSearch(s, true);
			}
			if (result) {
				html = rcxData.makeHtml(result);
				this.lastFound = [result];
			}
			else {
				html = '\u300C ' + s + ' \u300D was not found.';
				this.lastFound = [];
			}
			this.lastFound.fromLB = 1;

			let kanji = '';
			let have = {};
			let t = s + html;
			for (let i = 0; i < t.length; ++i) {
				let c = t.charCodeAt(i);
				if ((c >= 0x3000) && (c <= 0xFFFF)) {
					c = t.charAt(i);
					if (!have[c]) {
						result = rcxData.kanjiSearch(c);
						if (result) {
							this.lastFound.push(result);
							have[c] = 1;
							kanji += '<td class="q-k">' + rcxData.makeHtml(result) + '</td>';
						}
					}
				}
			}

			this.showPopup('<table class="q-tb"><tr><td class="q-w">' + html + '</td>' + kanji + '</tr></table>', null, null, true);
		}
	},

	lookupBoxKey: function(ev) {
		switch (ev.keyCode) {
		case 13:
			this.lookupSearch(ev.target.value);
			ev.stopPropagation();
			break;
		case 27:
			if (this.isVisible()) this.hidePopup();
			ev.target.value = "";
			ev.stopPropagation();
			break;
		}
	}
};


var rcxConfig = {
	observer: {
		observe: function(subject, topic, data) {
			// console.log('rcxConfig.observer: data=', data);
			if (data) {
				rcxConfigList.forEach(function(cfg) {
					if (cfg[1] == data) {
						rcxConfig.load();
						return;
					}
				});
				if (data == 'dpriority') {
					rcxConfig.load();
				}
			}
			else {
				rcxConfig.load();
			}
		}
	},

	addObserver: function() {
		this.prefBranch = Components.classes['@mozilla.org/preferences-service;1']
			.getService(Components.interfaces.nsIPrefService)
			.getBranch('extensions.rikaichan.');
		this.prefBranch.addObserver('', this.observer, false);
	},

	removeObserver: function() {
		this.prefBranch.removeObserver('', this.observer);
		this.prefBranch = null;
	},

	load: function() {
		// console.log('rcxConfig.load');
		let p = new rcxPrefs();

        let oldCss = rcxConfig.css;

		rcxConfigList.forEach(function(pair) {
			let [type, name] = pair;
			switch (type) {
			case 0:
				rcxConfig[name] = p.getInt(name);
				break;
			case 1:
				rcxConfig[name] = p.getString(name);
				break;
			case 2:
				rcxConfig[name] = p.getBool(name);
				break;
			}
		});


		let sss = Components.classes['@mozilla.org/content/style-sheet-service;1']
					.getService(Components.interfaces.nsIStyleSheetService);
		let ios = Components.classes['@mozilla.org/network/io-service;1']
					.getService(Components.interfaces.nsIIOService);
		if (rcxConfig.css != oldCss) {
			sss.loadAndRegisterSheet(ios.newURI('chrome://rikaichan/skin/popup-' + rcxConfig.css + '.css', null, null), sss.AUTHOR_SHEET);
			if (oldCss) sss.unregisterSheet(ios.newURI('chrome://rikaichan/skin/popup-' + oldCss + '.css', null, null), sss.AUTHOR_SHEET);
		}

		['cm', 'tm'].forEach(function(name) {
			let a = !rcxConfig[name + 'toggle'];
			let e = document.getElementById('rikaichan-toggle-' + name);
			if (e) e.hidden = a;

			let b = !rcxConfig[name + 'lbar'];
			e = document.getElementById('rikaichan-lbar-' + name);
			if (e) e.hidden = b;

			e = document.getElementById('rikaichan-separator-' + name);
			if (e) e.hidden = a || b;
		}, this);

		if (rcxMain.isTB) {
			rcxConfig.enmode = 1;
		}
		rcxMain.resendConfig();
		rcxData.loadConfig();
	}
};

rcxMain.init();
